import json
from typing import Dict


class JsonSerializable(object):
    @staticmethod
    def _from_json(json_dict: Dict, clazz):
        return clazz(**JsonSerializable._to_snake_case(json_dict))

    def to_json(self):
        return json.dumps(self, default=self._to_camel_case)

    def _to_camel_case(self, obj):
        return {JsonSerializable.snake_case_to_camel_case(k): v for k, v in obj.__dict__.items()}

    @staticmethod
    def _to_snake_case(json_dict: Dict):
        return {JsonSerializable.camel_case_to_snake_case(k): v for k, v in json_dict.items()}

    @staticmethod
    def snake_case_to_camel_case(s):
        parts = s.split('_')
        return parts[0] + ''.join(part.title() for part in parts[1:])

    @staticmethod
    def camel_case_to_snake_case(s):
        return ''.join(['_' + c.lower() if c.isupper() else c for c in s]).lstrip('_')



from typing import Callable, Any

from .receiver import ReceiverService
from .sender import SenderService
from ..domain import ClientConfig
from ..serializer import SerializationType


class StrmPrivacyClient(object):
    def __init__(self, client_id: str, client_secret: str, config: ClientConfig):
        """
        Class to interact with STRM Privacy. For each stream, a separate instance of `StrmPrivacyClient`
        is required.

        :param client_id: unique stream identifier
        :param client_secret: secret to authenticate this stream
        :param config: internal configuration (only change if instructed)
        """
        self._sender_service = SenderService(client_id, client_secret, config)
        self._receiver_service = ReceiverService(client_id, client_secret, config)

    async def start_timers(self):
        await self._sender_service.start_timer()
        await self._receiver_service.start_timer()

    async def send(self, event, serialization_type: SerializationType) -> str:
        return await self._sender_service.asend(event, serialization_type)

    async def start_receiving_ws(self, as_json: bool, consumer: Callable[[Any], Any]):
        return await self._receiver_service.start(as_json, consumer)

    async def close(self):
        self._receiver_service.close()



__author__ = """Stream Machine B.V."""
__email__ = 'apis@strmprivacy.io'
__version__ = '3.0.0'

from .client import StrmPrivacyClient
from .domain.config import ClientConfig
from .serializer import SerializationType
from .util import current_time_millis



import asyncio
import time
from contextlib import suppress


def current_time_millis():
    return int(round(time.time() * 1000))


class TimerTask(object):
    def __init__(self, method, time_in_seconds):
        """
        Timed repeating task

        :param method: the method to be invoked within this TimerTask
        :param time_in_seconds: frequency of invoking the method
        """
        self.method = method
        self.time_in_seconds = time_in_seconds
        self.running = False
        self._task = None

        self.method()

    async def start(self):
        if not self.running:
            self.running = True
            self._task = asyncio.ensure_future(self._run())

    async def stop(self):
        if self.running:
            self.running = False
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def _run(self):
        while True:
            await asyncio.sleep(self.time_in_seconds)
            self.method()



import datetime

import pytz
import requests
from requests import HTTPError

from ..domain.config import ClientConfig
from ..domain.base import JsonSerializable
from ..util import TimerTask


class AuthService(object):
    _headers = {"content-type": "application/x-www-form-urlencoded; charset=UTF-8"}

    def __init__(self, purpose: str, client_id: str, client_secret: str, config: ClientConfig):
        """
        Service responsible for ensuring the validity of access tokens for the StrmPrivacyClient

        :param purpose: for which is the AuthService initialized
        :param client_id: unique stream identifier
        :param client_secret: secret to authenticate this stream
        :param config: internal configuration
        """
        self._logger = config.get_logger(__name__)
        self._purpose = purpose
        self._client_id = client_id
        self._client_secret = client_secret
        self._config = config

        self.auth_provider: AuthProvider = None
        self.timer_task = TimerTask(self._initialize_auth_provider, config.auth_refresh_interval)

    async def start(self):
        await self.timer_task.start()

    def get_access_token(self) -> str:
        return self.auth_provider.access_token

    def get_refresh_token(self) -> str:
        return self.auth_provider.refresh_token

    def _authenticate(self, client_id: str, client_secret: str) -> None:
        self._logger.debug("authenticate")
        try:
            payload = f"grant_type=client_credentials&client_id={client_id}&client_secret={client_secret}"
            self._do_post(self._config.auth_auth_uri, payload)
        except HTTPError as e:
            self._logger.error(
                f"An error occurred while requesting an access token with clientId '{client_id}'")

    def _refresh(self, refresh_token: str, client_id: str, client_secret: str) -> None:
        self._logger.debug("_refresh")
        try:
            payload = f"grant_type=refresh_token&client_id={client_id}&" \
                      f"client_secret={client_secret}&refresh_token={refresh_token}"
            self._do_post(self._config.auth_auth_uri, payload)
        except HTTPError as e:
            self._logger.debug(f"Failed to refresh token with clientId '{client_id}'")
            self._logger.debug(
                f"Trying to request a new token with clientId '{client_id}'")

            self._authenticate(client_id, client_secret)

    def _do_post(self, uri: str, payload: str):
        response = None

        try:
            response = requests.post(uri, headers=AuthService._headers, data=payload)
            self.auth_provider = AuthProvider.from_json(response.json())
        except HTTPError as e:
            self._logger.error(e)
            self._logger.error(response.text)

            raise HTTPError(e, response.text)

    def _initialize_auth_provider(self):
        self._logger.debug("_initialize_auth_provider")
        if self.auth_provider is None:
            self._logger.debug(f"Initializing a new Auth Provider for {self._purpose}")
            self._authenticate(self._client_id, self._client_secret)
        elif self.auth_provider.is_almost_expired():
            self._logger.debug(f"Refreshing an existing Auth Provider {self._purpose}")
            self._refresh(self.auth_provider.refresh_token, self._client_id, self._client_secret)


class AuthProvider(JsonSerializable):
    _expiration_slack_time_seconds: int = datetime.timedelta(minutes=10).seconds

    def __init__(self, access_token: str, refresh_token: str, expires_in: int, **kwargs):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in
        self.expires_at = datetime.datetime.now(tz=pytz.utc).timestamp() + expires_in

    @staticmethod
    def from_json(json_dict: dict) -> 'AuthProvider':
        return JsonSerializable._from_json(json_dict, AuthProvider)

    def is_almost_expired(self):
        current_time = datetime.datetime.now(tz=pytz.utc).timestamp()

        return (current_time + AuthProvider._expiration_slack_time_seconds) >= self.expires_at


class AuthRequest(JsonSerializable):
    def __init__(self, client_id: str, client_secret: str):
        self.client_id = client_id
        self.client_secret = client_secret


class RefreshRequest(JsonSerializable):
    def __init__(self, refresh_token: str):
        self.refresh_token = refresh_token



from aiohttp import ClientSession

from .auth import AuthService
from ..domain import StrmPrivacyEventDTO
from ..domain.config import ClientConfig


class SenderService(object):
    def __init__(self, client_id: str, client_secret: str, config: ClientConfig):
        self._logger = config.get_logger(__name__)

        self.auth_service = AuthService(
            purpose=self.__class__.__name__,
            client_id=client_id,
            client_secret=client_secret,
            config=config
        )

        self._config = config
        self._session = ClientSession()
        self._requests_to_send = []

    async def start_timer(self):
        await self.auth_service.start()

    async def asend(self, event, serialization_type) -> str:
        dto = StrmPrivacyEventDTO(event, serialization_type)
        headers = {
            'Authorization': f'Bearer {self.auth_service.get_access_token()}',
            'Content-Type': 'application/octet-stream',
            'Strm-Serialization-Type': dto.get_serialization_type_header(),
            'Strm-Driver-Version': self._config.version.brief_string(),
            'Strm-Driver-Build': self._config.version.release_string(),
            'Strm-Schema-Ref': dto.get_schema_ref()
        }

        async with self._session.post(self._config.gateway_uri, data=dto.serialize(),
                                      headers=headers) as response:
            response_text = await response.text()

            if response.status != 204:
                self._logger.error(
                    f"Error while sending event to STRM Privacy ({self._config.gateway_uri}), response status = {response.status}, response: {response_text}")
            return response.status

    async def close(self):
        await self._session.close()




import dataclasses
import json
from abc import ABC, abstractmethod
from ctypes import Union
from io import BytesIO

import avro
from avro.io import DatumWriter
from avro.schema import Schema
from avro_json_serializer import AvroJsonSerializer
from jsonschema import validate

from .type import SerializationType, UnsupportedSerializationTypeException


class EventSerializer(ABC):
    @abstractmethod
    def serialize(self, event, serialization_type: SerializationType) -> bytes:
        pass


class AvroSerializer(EventSerializer):
    def __init__(self, schema: Schema):
        self._schema: Schema = schema
        self._serializer: Union[DatumWriter, AvroJsonSerializer] = None

    def serialize(self, event, serialization_type: SerializationType) -> bytes:
        if serialization_type is SerializationType.AVRO_BINARY:
            return self._serialize_avro_binary(event)
        elif serialization_type is SerializationType.AVRO_JSON:
            return self._serialize_avro_json(event)
        else:
            raise UnsupportedSerializationTypeException(
                f"The serialization type {serialization_type} is not supported for Avro")

    def _serialize_avro_binary(self, event_data: object) -> bytes:
        if self._serializer is None or not isinstance(self._serializer, DatumWriter):
            self._serializer = avro.io.DatumWriter(self._schema)

        bytes_writer = BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        self._serializer.write(event_data, encoder)
        bytes_writer.flush()

        return bytes_writer.getvalue()

    def _serialize_avro_json(self, event_data: object) -> bytes:
        if self._serializer is None or not isinstance(self._serializer, AvroJsonSerializer):
            self._serializer = AvroJsonSerializer(self._schema)

        return self._serializer.to_json(event_data).encode("utf-8")


class JsonSerializer(EventSerializer):
    # TODO add schema class type
    def __init__(self, schema: dict):
        self._schema = schema

    def serialize(self, event, _: SerializationType) -> str:
        validate(dataclasses.asdict(event), self._schema)

        return json.dumps(event, cls=JsonSerializer.Encoder)

    class Encoder(json.JSONEncoder):
        def default(self, o):
            if dataclasses.is_dataclass(o):
                return dataclasses.asdict(o)
            return super().default(o)



from .type import UnsupportedSerializationTypeException
from ...driver.serializer import AvroSerializer, EventSerializer, JsonSerializer


class SerializationProvider(object):
    _serializers = dict()

    def __init__(self):
        raise NotImplementedError(
            "SerializationProvider is not meant to be instantiated. Use the static methods instead.")

    @staticmethod  # TODO add schema class type
    def get_serializer(schema_ref: str, schema_type: str, schema) -> EventSerializer:
        existing_serializer = SerializationProvider._serializers.get(schema_ref)

        if existing_serializer is None:
            if schema_type == "avro":
                serializer = AvroSerializer(schema)

                SerializationProvider._serializers[schema_ref] = serializer

                return serializer
            elif schema_type == "json":
                serializer = JsonSerializer(schema)

                SerializationProvider._serializers[schema_ref] = serializer

                return serializer
            else:
                raise UnsupportedSerializationTypeException(f"Provided schema type '{schema_type}' is not supported")

        return existing_serializer



import logging

from pbr.version import SemanticVersion

from .. import __version__


class ClientConfig(object):
    def __init__(self, log_level=logging.INFO, **kwargs):
        self._log_level = log_level
        self._gateway_protocol = kwargs.get("gateway_protocol", "https")
        self._gateway_host = kwargs.get("gateway_host", "events.strmprivacy.io")
        self._gateway_endpoint = kwargs.get("gateway_endpoint", "/event")
        self._egress_protocol = kwargs.get("egress_protocol", "https")
        self._egress_host = kwargs.get("egress_host", "websocket.strmprivacy.io")
        self._egress_endpoint = kwargs.get("egress_endpoint", "/ws")
        self._egress_health_endpoint = kwargs.get("egress_health_endpoint", "/is-alive")
        self._auth_protocol = kwargs.get("auth_protocol", "https")
        self._auth_host = kwargs.get("auth_host", "accounts.strmprivacy.io")
        self._auth_endpoint = kwargs.get("auth_endpoint", "/auth/realms/streams/protocol/openid-connect/token")
        self._auth_refresh_interval = kwargs.get("auth_refresh_interval", 300)
        self._version = SemanticVersion.from_pip_string(__version__)

    def get_logger(self, name: str) -> logging.Logger:
        logger = logging.getLogger(name)
        logger.setLevel(self._log_level)
        return logger

    @property
    def gateway_uri(self):
        return f"{self._gateway_protocol}://{self._gateway_host}{self._gateway_endpoint}"

    @property
    def egress_uri(self):
        return f"{self._egress_protocol}://{self._egress_host}{self._egress_endpoint}"

    @property
    def egress_health_uri(self):
        return f"{self._egress_protocol}://{self._egress_host}{self._egress_health_endpoint}"

    @property
    def auth_auth_uri(self):
        return f"{self._auth_protocol}://{self._auth_host}{self._auth_endpoint}"

    @property
    def auth_refresh_interval(self):
        return self._auth_refresh_interval

    @property
    def version(self) -> SemanticVersion:
        return self._version



#!/usr/bin/env python
from setuptools import setup, find_namespace_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    author="Stream Machine B.V.",
    author_email='apis@strmprivacy.io',
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
    ],
    description="Python driver for interacting with STRM Privacy.",
    install_requires=[
        "requests>=2.24.0",
        "aiohttp>=3.6.2",
        "avro-python3>=1.10.0",
        "jsonschema>=3.2.0",
        "pytz>=2020.1",
        "pbr>=5.5.0",
        "six>=1.15.0",
        "responses>=0.10.15",
        "setuptools>=47.3.1",
        "pydantic>=1.5.1",
        "avro-json-serializer==1.0.3",
        "avro-python3>=1.10.0",
        "janus>=0.5.0"
    ],
    long_description=readme,
    include_package_data=True,
    keywords='strmprivacy api client driver',
    name='strmprivacy-driver',
    packages=find_namespace_packages(include=['strmprivacy.*']),
    namespace_packages=["strmprivacy"],
    setup_requires=[],
    version='3.0.0',
    zip_safe=False,
)



from typing import Callable, Any

import aiohttp
from aiohttp import ClientSession

from .auth import AuthService
from ..domain import ClientConfig


class ReceiverService(object):
    def __init__(self, client_id: str, client_secret: str, config: ClientConfig):
        self._logger = config.get_logger(__name__)

        self.auth_service = AuthService(
            purpose=self.__class__.__name__,
            client_id=client_id,
            client_secret=client_secret,
            config=config
        )

        self._config = config
        self._client = None
        self._running = False
        self._session = ClientSession()

    async def start_timer(self):
        await self.auth_service.start()

    async def start(self, as_json: bool, consumer: Callable[[Any], Any]):
        self._running = True

        await self.auth_service.start()
        uri = self._config.egress_uri + ("?asJson=true" if as_json else "")

        while self._running:
            async with self._session.ws_connect(uri, headers={
                'Authorization': f'Bearer {self.auth_service.get_access_token()}',
                'Strm-Driver-Version': self._config.version.brief_string(),
                'Strm-Driver-Build': self._config.version.release_string()
            }) as ws:
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        await consumer(msg.data)
                    elif msg.type == aiohttp.WSMsgType.CLOSED:
                        self._logger.debug("Websocket connection closed")
                        break
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        self._logger.debug("Error upon receiving data from websocket")
                        break

    def close(self):
        self._running = False



import asyncio
import logging
import time
from threading import Thread

import janus

from .. import StrmPrivacyClient, SerializationType
from ..domain.config import ClientConfig


class SyncSender(Thread):
    """
    A thread that owns an async event loop, and that receives events on a Janus queue from the synchronous
    part of the application. So this thing sends asynchronously, but can be called safely from multiple
    synchronous threads.

    If you use this class, `janus` needs to be added to the dependencies.
    """

    def __init__(self, client_id, client_secret, log_level=logging.DEBUG, async_debug=False):
        Thread.__init__(self)
        self._config = ClientConfig(log_level=logging.DEBUG)
        self._props = (client_id, client_secret)
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(log_level)
        self._queue = None
        self.async_debug = async_debug

    def run(self):
        """
        Start the event loop that handles the messages on the Janus queue
        :return:
        """
        asyncio.run(self.async_start(*self._props), debug=self.async_debug)

    async def async_start(self, client_id, client_secret):
        self._queue = janus.Queue()
        client = StrmPrivacyClient(client_id, client_secret, self._config)
        await client.start_timers()  # for the re-authorization jwt timer
        while True:
            event, serialization_type = await self._queue.async_q.get()
            await client.send(event, serialization_type)

    def wait_ready(self):
        """
        wait for the queue to become available
        :return:
        """
        while True:
            if self._queue is not None:
                return
            time.sleep(0.01)

    def send_event(self, event, serialization_type=SerializationType.AVRO_BINARY):
        self._queue.sync_q.put((event, serialization_type))



from dataclasses import dataclass


from ..serializer import SerializationType, SerializationProvider
from ..serializer.type import UnsupportedSerializationTypeException


@dataclass
class StrmPrivacyEventDTO(object):
    event: object
    serialization_type: SerializationType

    def get_schema_ref(self) -> str:
        return self.event.get_strm_schema_ref()

    def get_schema_type(self) -> str:
        return self.event.get_strm_schema_type()

    def get_serialization_type_header(self) -> str:
        if (self.serialization_type is SerializationType.JSON or
                self.serialization_type is SerializationType.AVRO_JSON):
            return "application/json"
        elif self.serialization_type is SerializationType.AVRO_BINARY:
            return "application/x-avro-binary"
        else:
            raise UnsupportedSerializationTypeException(
                f"Serialization type '{self.serialization_type}' is not supported")

    def serialize(self) -> bytes:
        stream_machine_schema = self.event.get_strm_schema()
        serializer = SerializationProvider.get_serializer(self.get_schema_ref(), self.get_schema_type(),
                                                          stream_machine_schema)

        return serializer.serialize(self.event, self.serialization_type)
